#include "extension.h"
#include <IBinTools.h>
#include <mathlib/vector.h>

FlashTools g_interface;		/**< Global singleton for extension's main interface */

SMEXT_LINK(&g_interface);

IGameConfig *g_pGameConf = NULL;
IBinTools *g_pBinTools = NULL;
IForward *g_pDeafen = NULL;
IForward *g_pPercentOfBlind = NULL;
IForward *g_pFBDetonate = NULL;
CDetour *DDeafen = NULL;
CDetour *DPercentOfBlind = NULL;
CDetour *DFBDetonate = NULL;

//Define DETOUR_DECL_STATIC3 since it dosnt exist in detours.h
#define DETOUR_DECL_STATIC3(name, ret, p1type, p1name, p2type, p2name, p3type, p3name) \
ret (*name##_Actual)(p1type, p2type, p3type) = NULL; \
ret name(p1type p1name, p2type p2name, p3type p3name)

DETOUR_DECL_STATIC3(PercentOfFlash, float, CBaseEntity *, entity1, Vector, pos, CBaseEntity *, entity2)
{
	int client = gamehelpers->EntityToBCompatRef(entity1);
	int index = gamehelpers->EntityToBCompatRef(entity2);
	float percent = DETOUR_STATIC_CALL(PercentOfFlash)(entity1, pos, entity2);
	float orgpercent = percent;

	cell_t vec[3];
	vec[0] = *(cell_t *)&pos.x;
	vec[1] = *(cell_t *)&pos.y;
	vec[2] = *(cell_t *)&pos.z;

	cell_t result = Pl_Continue;
	g_pPercentOfBlind->PushCell(client);
	g_pPercentOfBlind->PushCell(index);
	g_pPercentOfBlind->PushArray(vec, 3);
	g_pPercentOfBlind->PushFloatByRef(&percent);
	g_pPercentOfBlind->Execute(&result);

	if(result == Pl_Changed)
	{
		return percent;
	}
	if(result >= Pl_Handled)
	{
		return 0.0;
	}

	return orgpercent;
}
DETOUR_DECL_MEMBER1(DetourDeafen, void, float, distance)
{
	int client = gamehelpers->EntityToBCompatRef(reinterpret_cast<CBaseEntity *>(this));
	float pushdistance = distance;

	cell_t result = Pl_Continue;
	g_pDeafen->PushCell(client);
	g_pDeafen->PushFloatByRef(&pushdistance);
	g_pDeafen->Execute(&result);

	if(result == Pl_Continue)
	{
		DETOUR_MEMBER_CALL(DetourDeafen)(distance);
	}
	else if(result == Pl_Changed)
	{
		DETOUR_MEMBER_CALL(DetourDeafen)(pushdistance);
	}

	return;
}
DETOUR_DECL_MEMBER0(FlashDetonate, void)
{
	int index = gamehelpers->EntityToBCompatRef(reinterpret_cast<CFlashbangProjectile *>(this));

	cell_t result = Pl_Continue;
	g_pFBDetonate->PushCell(index);
	g_pFBDetonate->Execute(&result);

	if(result == Pl_Continue)
	{
		DETOUR_MEMBER_CALL(FlashDetonate)();
	}

	return;
}
bool FlashTools::SDK_OnLoad(char *error, size_t maxlength, bool late)
{
	char conf_error[255];
	if (!gameconfs->LoadGameConfigFile("fbtools.games", &g_pGameConf, conf_error, sizeof(conf_error)))
	{
		if (error)
		{
			snprintf(error, maxlength, "Could not read fbtools.games.txt: %s", conf_error);
		}
		return false;
	}

	sharesys->RegisterLibrary(myself, "flashtools");

	plsys->AddPluginsListener(this);

	CDetourManager::Init(g_pSM->GetScriptingEngine(), g_pGameConf);

	m_FBDetonateDetourEnabled = false;

	m_FlashPercentDetourEnabled = false;

	m_DeafenDetourEnabled = false;

	g_pDeafen = forwards->CreateForward("OnDeafen", ET_Event, 2, NULL, Param_Cell, Param_FloatByRef);
	g_pFBDetonate = forwards->CreateForward("OnFlashDetonate", ET_Event, 1, NULL, Param_Cell);
	g_pPercentOfBlind = forwards->CreateForward("OnGetPercentageOfFlashForPlayer", ET_Event, 4, NULL, Param_Cell, Param_Cell, Param_Array, Param_FloatByRef);

	return true;
}
bool FlashTools::SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlen, bool late)
{
	return true;
}
void FlashTools::SDK_OnUnload()
{
	gameconfs->CloseGameConfigFile(g_pGameConf);
	plsys->RemovePluginsListener(this);
	forwards->ReleaseForward(g_pDeafen);
	forwards->ReleaseForward(g_pFBDetonate);
	forwards->ReleaseForward(g_pPercentOfBlind);
}
void FlashTools::OnPluginLoaded(IPlugin *plugin)
{
	if(!m_FBDetonateDetourEnabled && g_pFBDetonate->GetFunctionCount())
	{
		m_FBDetonateDetourEnabled = CreateDetonateDetour();
	}
	if(!m_FlashPercentDetourEnabled && g_pPercentOfBlind->GetFunctionCount())
	{
		m_FlashPercentDetourEnabled = CreatePrecentOfBlindDetour();
	}
	if(!m_DeafenDetourEnabled && g_pDeafen->GetFunctionCount())
	{
		m_DeafenDetourEnabled = CreateDeafenDetour();
	}
}
void FlashTools::OnPluginUnloaded(IPlugin *plugin)
{
	if(m_FBDetonateDetourEnabled && !g_pFBDetonate->GetFunctionCount())
	{
		if(DFBDetonate != NULL)
		{
			DFBDetonate->Destroy();
			DFBDetonate = NULL;
		}
		m_FBDetonateDetourEnabled = false;
	}
	if(m_FlashPercentDetourEnabled && !g_pPercentOfBlind->GetFunctionCount())
	{
		if(DPercentOfBlind != NULL)
		{
			DPercentOfBlind->Destroy();
			DPercentOfBlind = NULL;
		}
		m_FlashPercentDetourEnabled = false;
	}
	if(m_DeafenDetourEnabled && !g_pDeafen->GetFunctionCount())
	{
		if(DDeafen != NULL)
		{
			DDeafen->Destroy();
			DDeafen = NULL;
		}
		m_DeafenDetourEnabled = false;
	}
}
bool CreateDeafenDetour()
{
	DDeafen =  DETOUR_CREATE_MEMBER(DetourDeafen, "Deafen");

	if(DDeafen != NULL)
	{
		DDeafen->EnableDetour();
		return true;
	}
	g_pSM->LogError(myself, "Deafen detour could not be initialized - Disabled OnDeafen forward");
	return false;
}
bool CreateDetonateDetour()
{
	DFBDetonate =  DETOUR_CREATE_MEMBER(FlashDetonate, "CFlashbangProjectile::Detonate");

	if(DFBDetonate != NULL)
	{
		DFBDetonate->EnableDetour();
		return true;
	}
	g_pSM->LogError(myself, "CFlashbangProjectile::Detonate detour could not be initialized - Disabled OnFlashDetonate forward");
	return false;
}
bool CreatePrecentOfBlindDetour()
{
	DPercentOfBlind =  DETOUR_CREATE_STATIC(PercentOfFlash, "PercentageOfFlashForPlayer");

	if(DPercentOfBlind != NULL)
	{
		DPercentOfBlind->EnableDetour();
		return true;
	}
	g_pSM->LogError(myself, "PercentageOfFlashForPlayer detour could not be initialized - Disabled OnGetPercentageOfFlashForPlayer forward");
	return false;
}
